/*
 *   stunnel       Universal SSL tunnel
 *   Copyright (c) 1998-2001 Michal Trojnara <Michal.Trojnara@mirt.net>
 *                 All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   In addition, as a special exception, Michal Trojnara gives
 *   permission to link the code of this program with the OpenSSL
 *   library (or with modified versions of OpenSSL that use the same
 *   license as OpenSSL), and distribute linked combinations including
 *   the two.  You must obey the GNU General Public License in all
 *   respects for all of the code used other than OpenSSL.  If you modify
 *   this file, you may extend this exception to your version of the
 *   file, but you are not obligated to do so.  If you do not wish to
 *   do so, delete this exception statement from your version.
 */

#include "common.h"
#include "prototypes.h"
#include "client.h"

/* protocol-specific function prototypes */
static int smb_client(CLI *);
static int smb_server(CLI *);
static int smtp_client(CLI *);
static int smtp_server(CLI *, int allow_downgrade);
static int pop3_client(CLI *);
static int pop3_server(CLI *, int allow_downgrade);
static int imap_client(CLI *);
static int imap_server(CLI *, int allow_downgrade);
static int nntp_client(CLI *);
static int nntp_server(CLI *);
static int telnet_client(CLI *);
static int telnet_server(CLI *);
static int RFC2487(int);

int negotiate(char *protocol, int client, CLI *c) {
    if(!protocol)
        return 0; /* No protocol negotiations */
    if(!c)
        return 0; /* No client present */
    log(LOG_DEBUG, "Negotiations for %s(%s side) started", protocol,
        client ? "client" : "server");
    if(!strcmp(protocol, "smb")) {
        if(client)
            return smb_client(c);
        else
            return smb_server(c);
    }
    if(!strncmp(protocol, "smtp",4)) {
        if(client)
            return smtp_client(c);
        else {
	  if (protocol[4] == '-')
            return smtp_server(c,1);
	  else if (protocol[4] == '\0')
	    return smtp_server(c,0);
	}
    }
    if(!strncmp(protocol, "pop3",4)) {
        if(client)
            return pop3_client(c);
        else {
	  if (protocol[4] == '-')
            return pop3_server(c,1);
	  else if (protocol[4] == '\0')
	    return pop3_server(c,0);
	}
    }
    if(!strncmp(protocol, "imap",4)) {
        if(client)
	  return imap_client(c);
        else {
	  if (protocol[4] == '-')
            return imap_server(c,1);
	  else if (protocol[4] == '\0')
	    return imap_server(c,0);
	}
    }
    if(!strcmp(protocol, "nntp")) {
        if(client)
            return nntp_client(c);
        else
            return nntp_server(c);
    }
    if(!strcmp(protocol, "telnet")) {
        if(client)
            return telnet_client(c);
        else
            return telnet_server(c);
    }
    log(LOG_ERR, "Protocol %s not supported in %s mode",
        protocol, client ? "client" : "server");
    return -1;
}

static int smb_client(CLI *c) {
    log(LOG_ERR, "Protocol not supported");
    return -1;
}

static int smb_server(CLI *c) {
    log(LOG_ERR, "Protocol not supported");
    return -1;
}

static int smtp_client(CLI *c) {
    char line[STRLEN];
    
    do { /* Copy multiline greeting */
        if(fdscanf(c->remote_fd, "%[^\n]", line)<0)
            return -1;
        if(fdprintf(c->local_wfd, "%s", line)<0)
            return -1;
    } while(strncmp(line,"220-",4)==0);

    if(fdprintf(c->remote_fd, "EHLO localhost")<0) /* Send an EHLO command */
        return -1;
    do { /* Skip multiline reply */
        if(fdscanf(c->remote_fd, "%[^\n]", line)<0)
            return -1;
    } while(strncmp(line,"250-",4)==0);
    if(strncmp(line,"250 ",4)!=0) { /* Error */
        log(LOG_ERR, "Remote server is not RFC 1425 compliant");
        return -1;
    }

    if(fdprintf(c->remote_fd, "STARTTLS")<0) /* Send STARTTLS command */
        return -1;
    do { /* Skip multiline reply */
        if(fdscanf(c->remote_fd, "%[^\n]", line)<0)
            return -1;
    } while(strncmp(line,"220-",4)==0);
    if(strncmp(line,"220 ",4)!=0) { /* Error */
        log(LOG_ERR, "Remote server is not RFC 2487 compliant");
        return -1;
    }
    return 0;
}

#define PP_BUFSIZE 8192
static int plaintext_proxy(CLI *c)
{
  fd_set rd_set, wr_set;
  char rbuf[PP_BUFSIZE];
  int rbufend,rdone;
  char lbuf[PP_BUFSIZE];
  int lbufend,ldone;
  struct timeval tv;
  int ready;
  int fdmax;
  int howmuch, sofar;

  fdmax = c->remote_fd;
  if (c->local_wfd > fdmax)
    fdmax = c->local_wfd;
  if (c->local_rfd > fdmax)
    fdmax = c->local_rfd;
  fdmax++;
  
  lbufend = rbufend = 0;
  rdone = ldone = 0;

  tv.tv_sec = 86400;
  tv.tv_usec = 0;
  while( !( (rdone || ldone) && (!rbufend && !lbufend) ) ) {
    log(LOG_DEBUG,"rdone=%d, ldone=%d, rbufend=%d, lbufend=%d",rdone,ldone,rbufend,lbufend);
    FD_ZERO(&rd_set);
    FD_ZERO(&wr_set);
    if (rbufend)
      FD_SET((c->local_wfd),&wr_set);
    else if (!rdone)
      FD_SET(c->remote_fd,&rd_set);
    
    if (lbufend)
      FD_SET(c->remote_fd,&wr_set);
    else if (!ldone)
      FD_SET(c->local_rfd,&rd_set);
    
    do { /* Skip "Interrupted system call" errors */
      ready=select(fdmax, &rd_set, &wr_set, NULL, &tv);
    } while(ready<0 && get_last_socket_error()==EINTR);
    
    if(ready<0) { /* Break the connection for others */
      sockerror("select");
      c->error=1;
      return 1;
    }
    if(!ready) { /* Timeout */
      log(LOG_DEBUG, "select timeout - connection reset");
      c->error=1;
      return 1;
    }
    
    if (lbufend > 0) {
      if (FD_ISSET(c->remote_fd,&wr_set))
	{
	  sofar = 0;
	  while (sofar < lbufend) {
	    if ((howmuch=write(c->remote_fd,lbuf+sofar,lbufend-sofar)) < 0) {
	      sockerror("write");
	      c->error=1;
	      return 1;
	    }
	    sofar += howmuch;
	  }
	  lbufend = 0;
	}
    } else {
      if (FD_ISSET(c->local_rfd,&rd_set)) {
	if ( (lbufend = read(c->local_rfd,lbuf,PP_BUFSIZE)) < 0) {
	  sockerror("read");
	  c->error = 1;
	  return 1;
	}
	if (lbufend == 0) /* EOF */
        {
	  log(LOG_DEBUG,"EOF reading from local");
	  ldone=1;
	}
      }
    }

    if (rbufend > 0) {
      if (FD_ISSET(c->local_wfd, &wr_set)) {
	sofar = 0;
	while (sofar < rbufend) {
	  if ( (howmuch = write(c->local_wfd, rbuf + sofar, rbufend - sofar)) < 0) {
	    sockerror("write");
	    c->error=1;
	    return 1;
	  }
	  sofar += howmuch;
	}
	rbufend = 0;
      }
    } else {
      if (FD_ISSET(c->remote_fd,&rd_set)) {
	if ( (rbufend = read(c->remote_fd,rbuf,PP_BUFSIZE)) < 0) {
	  sockerror("read");
	  c->error = 1;
	  return 1;
	}
	if (rbufend == 0) /* EOF */ {
	  log(LOG_DEBUG,"EOF reading from remote");
	  rdone=1;
	}
      }
    }
  }

  log(LOG_DEBUG, "plaintext_proxy finished normally.");
  return 0;
}

static int smtp_server(CLI *c, int allow_downgrade) {
    char line[STRLEN];
    int read_ret;

    if(RFC2487(c->local_rfd)==0)
        return 0; /* Return if RFC 2487 is not used */

    if(fdscanf(c->remote_fd, "220%[^\r\n]", line)!=1) {
        log(LOG_ERR, "Unknown server welcome");
        return -1;
    }
    if(fdprintf(c->local_wfd, "220%s + stunnel", line)<0)
        return -1;
    
    /* See if we get an EHLO command */
    if (((read_ret=fdgets(c->local_rfd, line)) >= 0) &&
	(strlen(line) >= 4) &&
	(!line[4] || isspace(line[4])) &&
	 (strncasecmp(line,"ehlo",4) == 0) ) {
      if (fdputs(c->remote_fd,line) < 0)
	return -1;
      while ( (fdgets(c->remote_fd, line) >= 0) 
	      && (line[3] == '-') ) {
	if (fdputs(c->local_wfd,line) < 0)
	  return -1;
      }
      line[3] = '-';
      if (fdputs(c->local_wfd,line) < 0)
	return -1;
      if (fdputs(c->local_wfd,"250 STARTTLS") < 0)
	return -1;
      
      /* See if we get a STARTTLS command */
      if (((read_ret=fdgets(c->local_rfd, line)) >= 0) &&
	  (strlen(line) >= 8) &&
	  (!line[8] || isspace(line[8])) &&
	  (strncasecmp(line,"starttls",8) == 0) ) {
	if (!line[8]) {
	  /* Technically, we should shut down the SMTP connection and get
	   * a new one, but screw it for now. */
	  if(fdputs(c->local_wfd, "220 Go ahead") >= 0)
	    return 0;
	} else {
	  fdputs(c->local_wfd,"501 Syntax error (no parameters allowed; STARTTLS disabled) (#5.5.4)");
	  return -1;
	} 
	
      }
    }
    if (read_ret < 0)
      return -1;
    if (allow_downgrade) {
      if (fdputs(c->remote_fd,line) < 0)
	return -1;
      exit(plaintext_proxy(c));
    }
    else {
      fdputs(c->local_wfd,"421 Encryption required; must use STARTTLS");
      return -1;
    }
}

static int pop3_client(CLI *c) {
    char line[STRLEN];

    if(fdscanf(c->remote_fd, "%[^\n]", line)<0)
        return -1;
    if(strncmp(line,"+OK ",4)) {
        log(LOG_ERR, "Unknown server welcome");
        return -1;
    }
    if(fdprintf(c->local_wfd, "%s", line)<0)
        return -1;
    if(fdprintf(c->remote_fd, "STLS")<0)
        return -1;
    if(fdscanf(c->remote_fd, "%[^\n]", line)<0)
        return -1;
    if(strncmp(line,"+OK ",4)) {
        log(LOG_ERR, "Server does not support TLS");
        return -1;
    }
    return 0;
}

static int pop3_server(CLI *c, int allow_downgrade) {
    char line[STRLEN];

    if(fdscanf(c->remote_fd, "%[^\r\n]", line)<0)
      return -1;

    if(fdprintf(c->local_wfd, "%s + stunnel", line)<0)
        return -1;

    while(1) {
      if(fdgets(c->local_rfd, line)<0)
        return -1;
      if(strncasecmp(line, "CAPA", 4) == 0) { /* Client wants RFC 2449 extensions */
	fdputs(c->remote_fd,line);
	if (fdgets(c->remote_fd,line) < 0)
	  return -1;
	if (line[0] == '+') {
	  if (fdputs(c->local_wfd,line) < 0)
	    return -1;
	  while ( (fdgets(c->remote_fd,line) >= 0)
		  && ( (strlen(line) > 1) || (line[0] != '.') ) ) {
	    if (fdputs(c->local_wfd,line) < 0)
	      return -1;
	  }
	} else {
	  if (fdputs(c->local_wfd,"+OK Stunnel capability list follows") < 0)
	    return -1;
	}
	  if (fdputs(c->local_wfd,"STLS") < 0)
	    return -1;
	  if (fdputs(c->local_wfd,".") < 0)
	    return -1;
      } else if (strncasecmp(line, "STLS", 4) == 0) {
	if(fdputs(c->local_wfd, "+OK stunnel starting TLS negotiation")<0)
	  return -1;
	return 0;
      }
      else {
	if (fdputs(c->remote_fd,line) < 0)
	  return -1;
	if (allow_downgrade)
	  exit(plaintext_proxy(c));
	else {
	  fdputs(c->local_wfd,"-Encryption required; must use STLS");
	  return -1;
	}
      }
    }
}

static int imap_client(CLI *c) {
    char line[STRLEN];

    if(fdgets(c->remote_fd, line)<0)
        return -1;
    if(strncasecmp(line,"* OK ",4)) {
        log(LOG_ERR, "Unknown server welcome");
        return -1;
    }
    if(fdputs(c->local_wfd, line)<0)
        return -1;
    if(fdprintf(c->remote_fd, "a STARTTLS")<0)
        return -1;
    if(fdgets(c->remote_fd, line)<0)
        return -1;
    if(strncasecmp(line,"a OK ",5)) {
        log(LOG_ERR, "Server does not support TLS");
        return -1;
    }
    return 0;
}

static int imap_server(CLI *c, int allow_downgrade) {
    char line[STRLEN];
    char procline[STRLEN];
    char *tag, *cmd;
    int do_if_fail = 0;

    if (fdgets(c->remote_fd,line) < 0)
      return -1;

    if(fdprintf(c->local_wfd, "%s + stunnel", line)<0)
        return -1;

    while(1) {
      if(fdgets(c->local_rfd, line)<0)
        return -1;
      safecopy(procline, line);
      tag = procline;
      if (!(cmd=strchr(procline,' '))) {
	do_if_fail = 1;
	break;
      }
      *cmd='\0';
      cmd++;
      if(strncasecmp(cmd, "CAPABILITY", 11) == 0) {
	fdputs(c->remote_fd,line);
	if (fdgets(c->remote_fd,line) < 0)
	  return -1;
	if (strncasecmp(line,"* CAPABILITY",12) == 0) {
	  safeconcat(line," STARTTLS");
	  fdputs(c->local_wfd,line);
	} else {
	  do_if_fail = 2;
	  break;
	}
	if (fdgets(c->remote_fd,line) < 0)
	  return -1;
	if ( (strncasecmp(line,tag,strlen(tag)) != 0) ||
	     (line[strlen(tag)] != ' ') ||
	     (strncasecmp(line+strlen(tag)+1,"OK",2) != 0) )
	  {
	    do_if_fail = 2;
	    break;
	  }
	if (fdputs(c->local_wfd,line) < 0)
	  return -1;
      }
      else if (strncasecmp(cmd, "STARTTLS", 8) == 0) {
	if (fdwrite(c->local_wfd,tag,strlen(tag)) < 0)
	  return -1;
	if (fdwrite(c->local_wfd," ",1) < 0)
	  return -1;
	if (fdputs(c->local_wfd,"OK Begin TLS negotiation now") < 0)
	  return -1;

	return 0;
      } else {
	do_if_fail = 1;
	break;
      }
    }
    
    if (allow_downgrade) {
      if (do_if_fail == 1) {
	if (fdputs(c->remote_fd,line) < 0)
	  return -1;
      } else if (do_if_fail == 2) {
	if (fdputs(c->local_wfd,line) < 0)
	  return -1;
      }
      exit(plaintext_proxy(c));
    }
    else {
      fdputs(c->local_wfd,"-Encryption required; must use STLS");
      return -1;
    }

}


static int nntp_client(CLI *c) {
    char line[STRLEN];

    if(fdscanf(c->remote_fd, "%[^\n]", line)<0)
        return -1;
    if(strncmp(line,"200 ",4) && strncmp(line,"201 ",4)) {
        log(LOG_ERR, "Unknown server welcome");
        return -1;
    }
    if(fdprintf(c->local_wfd, "%s", line)<0)
        return -1;
    if(fdprintf(c->remote_fd, "STARTTLS")<0)
        return -1;
    if(fdscanf(c->remote_fd, "%[^\n]", line)<0)
        return -1;
    if(strncmp(line,"382 ",4)) {
        log(LOG_ERR, "Server does not support TLS");
        return -1;
    }
    return 0;
}

static int nntp_server(CLI *c) {
    log(LOG_ERR, "Protocol not supported in server mode");
    return -1;
}

static int telnet_client(CLI *c) {
    log(LOG_ERR, "Protocol not supported");
    return -1;
}

static int telnet_server(CLI *c) {
    log(LOG_ERR, "Protocol not supported");
    return -1;
}

static int RFC2487(int fd) {
    fd_set         fdsRead;
    struct timeval timeout;

    FD_ZERO(&fdsRead);
    FD_SET(fd, &fdsRead);
    timeout.tv_sec=timeout.tv_usec=0; /* don't wait */

    switch(select(fd+1, &fdsRead, NULL, NULL, &timeout)) {
    case 0: /* fd not ready to read */
        log(LOG_DEBUG, "RFC 2487 detected");
        return 1;
    case 1: /* fd ready to read */
        log(LOG_DEBUG, "RFC 2487 not detected");
        return 0;
    }
    sockerror("RFC2487 (select)");
    return -1;
}

/* End of protocol.c */
